# The skeleton for this code was taken and adapted from https://github.com/Ultimaker/UraniumExampleToolPlugin


# Copyright (c) 2017 Ultimaker B.V.
# This example is released under the terms of the AGPLv3 or higher.

import os.path #To find the QML files.
from PyQt5.QtCore import pyqtProperty, pyqtSignal, Qt, QUrl, QObject, QVariant, pyqtSlot #To define a shortcut key and to find the QML files, and to expose information to QML.
from PyQt5.QtQml import QQmlComponent, QQmlContext #To create a dialogue window.

from UM.Application import Application #To register the information dialogue.
from UM.Event import Event, MouseEvent #To understand what events to react to.
from UM.PluginRegistry import PluginRegistry #To find the QML files in the plug-in folder.
from UM.Scene.Selection import Selection #To get the current selection and some information about it.
from UM.Scene.Camera import Camera
from UM.Scene.SceneNode import SceneNode
from UM.View.Renderer import Renderer
from UM.Tool import Tool #The PluginObject we're going to extend.
from UM.Mesh.MeshBuilder import MeshBuilder
from UM.Mesh.MeshData import MeshData, calculateNormalsFromIndexedVertices
from UM.Message import Message
from UM.Qt.Bindings.MainWindow import MainWindow

from UM.Math.Vector import Vector
from UM.Math.Matrix import Matrix
from UM.Math.Plane import Plane
from UM.Math.Ray import Ray
from UM.Math.Float import Float
from UM.Math.Color import Color
from UM.Math.AxisAlignedBox import AxisAlignedBox

from UM.Operations.AddSceneNodeOperation import AddSceneNodeOperation
from UM.Operations.RemoveSceneNodeOperation import RemoveSceneNodeOperation
from UM.Operations.TranslateOperation import TranslateOperation
from UM.Operations.GroupedOperation import GroupedOperation

from cura.Scene.CuraSceneNode import CuraSceneNode
from cura.Scene.SliceableObjectDecorator import SliceableObjectDecorator
from cura.Scene.BuildPlateDecorator import BuildPlateDecorator

import numpy
import trimesh
import sys
import time
import threading
from copy import deepcopy
from collections import OrderedDict

from .tripy import earclip #from https://github.com/linuxlewis/tripy/blob/master/tripy.py

#from .CuttingOperation import CuttingOperation
from .SetTransformMatrixOperation import SetTransformMatrixOperation #from https://github.com/fieldOfView/Cura-MeshTools
from .SetParentOperationSimplified import SetParentOperationSimplified #from https://github.com/fieldOfView/Cura-MeshTools
from .SetMeshDataAndNameOperation import SetMeshDataAndNameOperation #from https://github.com/fieldOfView/Cura-MeshTools



############################################################################################################
# Main Code #
############################################################################################################


# The class for our cutting tool, automatically creates a button in the toolbar on left edge of Cura
# also inherits from QObject for QML UI component (not actually used at the moment)
class ExampleTool(QObject, Tool):
    # Creates an instance of this tool.
    def __init__(self, parent = None) -> None:
        QObject.__init__(self, parent)
        Tool.__init__(self)

        #self._shortcut_key = Qt.Key_X

        # once a cutting area has been defined the plugin creates a message on bottom screen with two buttons
        # for confirming or aborting the cut (Cura-internal message class)

        # _message is the message for confirming/aborting cuts, confirmCut is called when one of the buttons is pressed
        self._message = Message(title="Cutting Tool", lifetime=0, dismissable=False)
        self._message.setText("Would you like to confirm the cut?")
        self._message.addAction(action_id="Abort", name="Abort", icon="knife2.svg", description="Abort Cut")
        self._message.addAction(action_id="Confirm", name="Confirm", icon="knife2.svg", description="Confirm Cut")
        self._message.actionTriggered.connect(self.confirmCut)
        self._message.actionTriggered.connect(self.abortCut)
        #this is probably illegal but we want to force the user to react to the message through the buttons
        self._message._dismissable = False


        # _error is a regular message (no buttons except standard dismiss) to diplay error messages
        # its error text is set according to the error once it has occurred
        self._error = Message(title="An Error Occurred", lifetime=10)

        # this message will be shown when a cut splits the selected model into more than two disconnected bodies
        # the user has to decide whether they want to have all these bodies as separate objects/scenenodes
        # or not (in which case cutting will result in two objects, with multiple connectivity components)
        self._moreThanTwo = Message(title="Cutting Tool", lifetime=0, dismissable=False)
        self._moreThanTwo.setText("The defined plane splits the model into more than two bodies.\nDo you wish all these bodies to become independent objects?")
        self._moreThanTwo.addAction(action_id="onlyTwo", name="No", icon="knife2.svg", description="No")
        self._moreThanTwo.addAction(action_id="moreThanTwo", name="Yes", icon="knife2.svg", description="Yes")
        self._moreThanTwo.actionTriggered.connect(self._moreThanTwoHandler)
        #this is probably illegal but we want to force the user to react to the message through the buttons
        self._moreThanTwo._dismissable = False

        self._wait_message = Message(title="Cutting Now Underway", lifetime=0, use_inactivity_timer=False)
        self._wait_message.setText("Please have a little patience while your model is being cut.")
        self._wait_message.setProgress(-1)
        self._wait_message._inactivity_timer = None
        Application.getInstance().hideMessageSignal.connect(self.showWait)
        self._finish_message = Message(title="Cutting Finished", lifetime=10)
        self._finish_message.setText("The cutting process was completed successfully. Thank you for waiting.")

        # helper variables
        self._dragging = False              #True when a mouse drag for defining cutting plane is underway
        self._startVertex = Vector(0,0,0)   #for remembering starting point of mouse drag
        self._endVertex = Vector(0,0,0)     #for remembering ending point of mouse drag
        self._camRay = None                 #for remembering viewing direction (for expanding dragvector into square)
        self._cutting = False               #True when a cutting plane has been defined but cutting operation not started yet
        self._cutOrg = None                 #for remembering origin vector of cutting plane
        self._cutNrm = None                 #for remembering normal vector of cutting plane
        self._cutSN = None                  #for remembering SceneNode representing cutting plane
        self.newMesh1 = None
        self.newMesh2 = None
        self.selectedModel = None

        self._DeactivateTool_message = Message(title="Tool Deactivating", lifetime=10)
        self._DeactivateTool_message.setText("You are Deactivating the cutting tool. Select any tool from the left to restart")

        self._myselectedModel = None # initialize it

    # During development there was an issue, that _wait_message would disappear earlier than it should
    # This function catches the hiding of any messages and returns _wait_message to a visible state in those unwanted cases
    def showWait(self, id):
        if id is self._wait_message and self._cutSN is not None :
            self._wait_message.show()


    ##  Called when something happens in the scene while our tool is active.
    #
    #   For instance, we can react to mouse clicks, mouse movements, and so on.
    def event(self, event):
        Tool.event(self, event) #The super event updates the selection if the user clicks on some object.
        active_cam = Application.getInstance().getController().getScene().getActiveCamera() #get the current camera

        self_Controller = Application.getInstance().getController()
        self_Controller.setFallbackTool("ExampleTool")

        # We only react to events involving the left mouse button (press, drag, release)

        # define first vertex/origin of cutting plane (only applies when mouse drag follows)
        if event.type == Event.MousePressEvent:
            if MouseEvent.LeftButton not in event.buttons:
                return False

            if self._cutting:
                return False

            cam_ray = active_cam.getRay(0,0)
            cam_dir = cam_ray.direction

            # set the drag plane with normal vector the camera viewing direction
            self.setDragPlane(Plane(cam_dir, 0.0))
            self.setDragStart(event.x, event.y)
            self._dragging = False
            # the drag start will be one of the four verteces of rect cutting area
            self._startVertex = self.getDragStart()

            self_Controller.setCameraTool("CameraTool")
            self_Controller.setSelectionTool("SelectionTool")
            self._myselectedModel = None
            return True

        # Drag to second vertex in cutting plane to define its tilt (from current pov)
        if event.type == Event.MouseMoveEvent:
            if MouseEvent.LeftButton not in event.buttons:
                return False

            if self._cutting:
                return False

            if not self.getDragPlane():
                return False

            if not self.getDragStart():
                self.setDragStart(event.x, event.y)
                if not self.getDragStart(): #May have set it to None.
                    return False

            # when first dragging, a SceneNode to visualize and remember the defined cutting area is initialized
            if not self._dragging:
                self._dragging = True
                self._cutSN = SceneNode()
                self._cutSN.setVisible(True)
                self._cutSN.setSelectable(True)
                self._cutSN.setParent(Application.getInstance().getController().getScene().getRoot())

            # make cutting area already visible while dragging
            self._endVertex = self.getDragPosition(event.x, event.y)
            self._camRay = active_cam.getRay(0.0, 0.0)
            self.drawCuttingArea()

            return True

        # when releasing mouse button the cutting area is defined and cutting can commence when user chooses to
        # user can still change cutting area after initial definition by dragging
        if event.type == Event.MouseReleaseEvent:
            if MouseEvent.LeftButton not in event.buttons:
                return False

            if not self._dragging:
                return False

            if self._dragging:
                self._dragging = False
                # the point where dragging stops will be second vertex of cutting area
                self._endVertex = self.getDragPosition(event.x, event.y)
                self._camRay = active_cam.getRay(0.0, 0.0)

                self._cutting = True

                # build cutting area from two verteces and camera ray, render into scene
                self.drawCuttingArea()
                # show message with buttons for confirming/aborting cutting operation
                self._finish_message.hide()
                self._error.hide()
                self._message.show()

            self.setDragPlane(None)
            self._startVertex = Vector(0,0,0)
            self._endVertex = Vector(0,0,0)
            self._camRay = None

            self.propertyChanged.emit()

            self_Controller.setCameraTool("CameraTool")
            self_Controller.setSelectionTool("SelectionTool")
            self._myselectedModel = None
            return True

        if event.type == Event.ToolActivateEvent:
            self_Controller.setCameraTool("CameraTool")
            self_Controller.setSelectionTool("SelectionTool")

            if Selection.getCount() > 0:
                if not Selection.isSelected(self._cutSN):
                    if self._cutSN is None: # if at the moment the cutting plane is not yet created
                        #Cura SceneNode
                        self._myselectedModel = Selection.getSelectedObject(0) #save the object for later clicking outside cases
                        self_Controller.setCameraTool("ExampleTool")
                        self_Controller.setSelectionTool("ExampleTool")

                if Selection.isSelected(self._cutSN):
                    self._myselectedModel = None

            return True

        if event.type == Event.ToolDeactivateEvent:

            if Selection.getCount() == 0:   # above, after selecting the cutting plane, self._myselectedModel is set to None

                if self._myselectedModel is not None: # make sure that there's at least one SceneNode

                    Selection.add(self._myselectedModel)
                    self_Controller.setActiveTool("ExampleTool")
                    self._DeactivateTool_message.show()

            return True


    # this is called when either button in _moreThanTwo is pressed
    def _moreThanTwoHandler(self, id, name):
        self._error.hide()
        self._wait_message.hide()
        self._message.hide()
        self._moreThanTwo.hide()

        if self.selectedModel is not None:
            if name == "moreThanTwo":
                l = numpy.append(self.newMesh1.split(only_watertight=False), self.newMesh2.split(only_watertight=False), axis=0)
                self._replaceSceneNode(self.selectedModel, l)
            else:
                self._replaceSceneNode(self.selectedModel, [self.newMesh1, self.newMesh2])

            if not self.newMesh1.is_watertight or not self.newMesh2.is_watertight:
                self._error.setText("The cutting process did not yield watertight meshes.")
                self._error.show()
            else:
                self._finish_message.show()

        self._cutting = False
        self._cutOrg = None
        self._cutNrm = None
        self._cutSN.setParent(None)
        if Selection.isSelected(self._cutSN):
            Selection.remove(self._cutSN)
        self._cutSN = None
        self.newMesh1 = None
        self.newMesh2 = None
        self.selectedModel = None


    # this is called when either button in _message is pressed
    def confirmCut(self, id, name):
        self._message.hide()

        if not self._cutting:
            return

        # if the confirm button was pressed, the model is cut along the defined (possibly tweaked) plane
        if name == "Confirm":
            # currently only works for a single selected model
            if len(Selection.getAllSelectedObjects()) < 1:
                self._wait_message.hide()
                self._finish_message.hide()
                self._error.hide()
                self._error.setText("Please select one printable Cura Object to cut.")
                self._message.show()
                self._error.show()
                return
            if len(Selection.getAllSelectedObjects()) > 1:
                self._wait_message.hide()
                self._finish_message.hide()
                self._error.hide()
                self._error.setText("Cutting several models at the same time is not supported.")
                self._message.show()
                self._error.show()
                return
            self.selectedModel = Selection.getSelectedObject(0)
            if self.selectedModel is None:
                self._cutting = False
                self._cutOrg = None
                self._cutNrm = None
                self._cutSN.setParent(None)
                if Selection.isSelected(self._cutSN):
                    Selection.remove(self._cutSN)
                self._cutSN = None
                return

            if not isinstance(self.selectedModel, CuraSceneNode):
                self._wait_message.hide()
                self._finish_message.hide()
                self._error.hide()
                self._error.setText("Please select one printable Cura Object to cut.")
                self._message.show()
                self._error.show()
                self.selectedModel = None
                return

            # build the new meshdata for the two model-sections
            self._finish_message.hide()
            self._error.hide()
            self._message.hide()
            self._wait_message.show()

            thr = threading.Thread(target=self.splitMesh, args=(self.selectedModel,))
            thr.start()

        else:
            self._wait_message.hide()
            # remove cutting area whether cut was confirmed or not
            self._cutting = False
            self._cutOrg = None
            self._cutNrm = None
            self._cutSN.setParent(None)
            if Selection.isSelected(self._cutSN):
                Selection.remove(self._cutSN)
            self._cutSN = None
            self.newMesh1 = None
            self.newMesh2 = None

    @pyqtSlot(str, str)
    def abortCut(self, id, name):
        if name == "Abort":
            # default
            Application.getInstance().getController().setCameraTool("CameraTool")
            Application.getInstance().getController().setSelectionTool("SelectionTool")

            if not Selection.isSelected(self._cutSN) and Selection.getCount() > 0: # when sth is selected, but not the cutting plane
                Application.getInstance().getController().setCameraTool("ExampleTool")
                Application.getInstance().getController().setSelectionTool("ExampleTool")

############################################################################################################
# Helper and Maths Functions #
############################################################################################################


    # the meat of the cutting operation code
    # the given SceneNode (model) is spit along the defined cutting plane, and for the two resuling sections
    #
    def splitMesh(self, model: SceneNode):
        triModel = self._toTriMesh(model.getMeshData())
        cutMesh = self._toTriMesh(self._cutSN.getMeshData())

        verts = self._cutSN.getMeshDataTransformedVertices()
        orientation = model.getWorldOrientation().toMatrix().getData()
        #we have to bring the cutting area into the model's mesh's local coordinate space
        posDifference = verts[0] - model.getWorldPosition().getData()
        org = posDifference
        org = numpy.dot(numpy.append(org, [1.0]), orientation)
        org = org[0:-1]
        org = numpy.divide(org, model.getWorldScale().getData())
        self._cutOrg = org

        #calculate normal vector of cutting plane as cross product of spanning vectors
        nrm = numpy.cross(numpy.subtract(verts[3], verts[2]), numpy.subtract(verts[1], verts[2]))
        nrm = numpy.dot(numpy.append(nrm, [1.0]), orientation)
        nrm = nrm[0:-1]
        self._cutNrm = nrm
        nrmOpp = nrm * (-1)

        # slice_plane splits trimesh along given plane, automatically creates new vertices + edges along plane,
        # but the cross-section is left open -> has to be closed off by flat surface manually
        newTriModel1 = triModel.slice_plane(org, nrm)
        newTriModel2 = triModel.slice_plane(org, nrmOpp)
        newTriModel1.merge_vertices()
        newTriModel2.merge_vertices()


        tempVerts1 = newTriModel1.vertices
        tempFaces1 = newTriModel1.faces
        tempVerts2 = newTriModel2.vertices
        tempFaces2 = newTriModel2.faces

        # in case cutting splits model into more than two sections, we triangulate the cross-sections for
        # all resulting model-sections seperately
        connectivityComps1 = newTriModel1.split(only_watertight=False)
        connectivityComps2 = newTriModel2.split(only_watertight=False)

        if isinstance(connectivityComps1, trimesh.base.Trimesh):
            connectivityComps1 = [connectivityComps1]
        if isinstance(connectivityComps2, trimesh.base.Trimesh):
            connectivityComps2 = [connectivityComps2]

        if len(connectivityComps1) == 0 or len(connectivityComps1) == 0:
            self._wait_message.hide()
            self._finish_message.hide()
            self._error.setText("The specified plane does not cut the model.")
            self._error.show()
            return None, None

        if len(connectivityComps2) > len(connectivityComps1):
            connectivityComps = connectivityComps2
        else:
            connectivityComps = connectivityComps1

        for c in range(len(connectivityComps)):

            # prepare boolean masks to isolate those vertices from the mesh which were newly created by slice_plane
            # these are the cornerpoints (all in one plane) of the polygon cross-section, which has to be closed
            #newTriModel1.remove_degenerate_faces()
            #self.removeStupidFaces(newTriModel1)
            #newTriModel1.remove_unreferenced_vertices()
            cutVerts = self.getIntersectingVertices(triModel, connectivityComps[c])
            temp = connectivityComps[c].vertices.tolist()
            mask1 = numpy.zeros(len(temp), dtype=bool)
            for i in range(len(temp)):
                if temp[i] in cutVerts.tolist():
                    mask1[i] = True

            # as of now, we compute the polygon mesh closing the cross-section of the new submodels as the convex hull
            # of the cornerpoints from above
            # we then merge each of the sections of the original models with one cross-section mesh, resulting
            # in two seperate, (hopefully) watertight, meshes for the two sections
            tempTriModel1 = connectivityComps[c].copy()
            tempTriModel1.update_vertices(mask1)
            if numpy.shape(tempTriModel1.vertices)[0] == 0:
                continue

            self.removeStupidFaces(tempTriModel1)

            tTriModelList = self.triangulateIntersectingPolygon(connectivityComps[c], tempTriModel1)
            if tTriModelList is None:
                self._wait_message.hide()
                self._finish_message.hide()
                self._error.show()
                return None, None

            for t in list(range(len(tTriModelList))):
                tTriModel1 = tTriModelList[t]

                tempFaces1 = numpy.append(tempFaces1, tTriModel1.faces + numpy.shape(tempVerts1)[0], axis=0)
                tempVerts1 = numpy.append(tempVerts1, tTriModel1.vertices, axis=0)

                tempFaces2 = numpy.append(tempFaces2, tTriModel1.faces + numpy.shape(tempVerts2)[0], axis=0)
                tempVerts2 = numpy.append(tempVerts2, tTriModel1.vertices, axis=0)

        TriModel1 = trimesh.base.Trimesh(vertices=tempVerts1, faces=tempFaces1, visual=newTriModel1.visual)
        TriModel1.merge_vertices()
        TriModel1.fill_holes()
        self.removeStupidFaces(TriModel1)

        TriModel2 = trimesh.base.Trimesh(vertices=tempVerts2, faces=tempFaces2, visual=newTriModel2.visual)
        TriModel2.merge_vertices()
        TriModel2.fill_holes()
        self.removeStupidFaces(TriModel2)

        self.newMesh1 =  TriModel1
        self.newMesh2 =  TriModel2


        if self.newMesh1 is None or self.newMesh2 is None:
            self._cutting = False
            self._cutOrg = None
            self._cutNrm = None
            self._cutSN.setParent(None)
            if Selection.isSelected(self._cutSN):
                Selection.remove(self._cutSN)
            self._cutSN = None
            self._wait_message.hide()
            return

        # depending on the number of resulting connectivity components, either _moreThanTwo message is shown
        # so the user can decide whether he wants a new object for each, or _moreThanTwoHandler is called
        # directly to make two new ojects

        if self.newMesh1.body_count > 1 or self.newMesh2.body_count > 1:
            self._finish_message.hide()
            self._error.hide()
            self._message.hide()
            self._wait_message.hide()
            self._moreThanTwo.show()
            return
        else:
            self._moreThanTwoHandler("onlyTwo", "onlyTwo")
            self._wait_message.hide()
            return


    # given the trimesh of the outline of a model section after cutting, this function triangulates
    def triangulateIntersectingPolygon(self, openMesh: trimesh.base.Trimesh, sectionMesh: trimesh.base.Trimesh):
        verts = sectionMesh.vertices
        edgesList = self.getIntersectingEdges(openMesh, sectionMesh)
        if edgesList is None:
            self._error.setText("There was a problem with the cutting process. (Error in Edges)")
            return None
        #edges = edges.tolist()
        A, B, C = self.getNonCollinearVertices(verts)
        if A is None or B is None or C is None:
            self._error.setText("There was a problem with the cutting process. (Error in NonCollinearVertices)")
            return None
        verts = verts.tolist()
        M, Minv = self.calc3Dto2DTransformMatrix(A, B, C)
        tVerts = self.transformToOtherDimension(sectionMesh.vertices, M).tolist()

        trianglesList = []

        for e in range(len(edgesList)):
            edges = edgesList[e].tolist()
            trianglesIndices = []

            sortedVerts = [verts[edges[0][0]]]
            sortedVertsIndices = [edges[0][0]]
            for j in range(0, len(verts)):
                prior = sortedVertsIndices[-1]
                for i in range(0, len(verts)):
                    if [prior, i] in edges and i not in sortedVertsIndices:
                        sortedVerts.append(verts[i])
                        sortedVertsIndices.append(i)
                        break

            # prepare vertices in format required by tripy.earclip
            twodVerts = self.transformToOtherDimension(numpy.array(sortedVerts), M)
            polygon = []
            for i in range(numpy.shape(twodVerts)[0]):
                polygon.append((twodVerts[i][0], twodVerts[i][1]))
            triangles = earclip(polygon)

            for i in range(len(triangles)):
                currTriangle = [list(l) for l in triangles[i]]
                trianglesIndices.append([tVerts.index(currTriangle[j]) for j in range(3)])

            trianglesList.append(trimesh.base.Trimesh(vertices=sectionMesh.vertices, faces=numpy.array(trianglesIndices), visual=sectionMesh.visual))
            #numpy.take(sectionMesh.vertices, sortedVertsIndices, 0)

        return trianglesList

    # update_vertices leaves a lot of crap in the mesh's faces, which we remove here
    def removeStupidFaces(self, mesh: trimesh.base.Trimesh):
        faces = mesh.faces.tolist()
        mask = numpy.zeros(len(faces), dtype=bool)

        for i in range(len(faces)):
            f = list(OrderedDict.fromkeys(faces[i]))
            if len(f) == 3:
                mask[i] = True

        mesh.update_faces(mask)

    # calculate which vertices in a section of a split up model lie within the cutting plane
    # (i.e. they were newly created during the cutting operation)
    # these vertices are the endpoints of the outline edges of the section-body
    def getIntersectingVertices(self, openMesh: trimesh.base.Trimesh, sectionMesh: trimesh.base.Trimesh):

        sectionVerts = sectionMesh.vertices.tolist()
        intersectingVertices = []
        for i in range(len(sectionVerts)):
            if self.isVertexInPlane(sectionVerts[i], self._cutOrg, self._cutNrm):
                intersectingVertices.append(i)

        result = numpy.take(sectionMesh.vertices, intersectingVertices, 0)

        return result

    # calculate which edges in a section of a split up model run along (i.e. lie within) the cutting plane
    # (also known as outline of a non-watertight 3D body)
    # this is achieved by taking the edges which only belong to one face of the body (two is normal in watertight bodies)
    def getIntersectingEdges(self, openMesh: trimesh.base.Trimesh, sectionMesh: trimesh.base.Trimesh):
        edges = openMesh.edges.tolist()
        openVerts = openMesh.vertices.tolist()
        outlineEdges =[]
        cutEdges = []
        verts = self.getIntersectingVertices(openMesh, sectionMesh).tolist()
        remainingVerts = deepcopy(verts)
        result = []
        edgeCount = 0

        for ii in range(len(verts)):
            i = openVerts.index(verts[ii])
            for jj in range(len(verts)):
                j = openVerts.index(verts[jj])
                if [i,j] in edges and [j,i] not in edges:
                    if [ii,jj] not in outlineEdges and [jj,ii] not in outlineEdges:
                        outlineEdges.append([ii,jj])
                        outlineEdges.append([jj,ii])
                if [j,i] in edges and [i,j] not in edges:
                    if [jj,ii] not in outlineEdges and [ii,jj] not in outlineEdges:
                        outlineEdges.append([jj,ii])
                        outlineEdges.append([ii,jj])

        while len(remainingVerts) != 0:
            cutEdges = []
            sortedVerts = [remainingVerts[0]]
            sortedVertsIndices = [0]
            del remainingVerts[0]
            for j in range(1, len(verts)):
                prior = sortedVerts[-1]
                for i in range(1, len(verts)):
                    if [verts.index(prior), i] in outlineEdges and [verts.index(prior), i] not in cutEdges and [i, verts.index(prior)] not in cutEdges:
                        sortedVerts.append(verts[i])
                        sortedVertsIndices.append(i)
                        cutEdges.append([verts.index(prior), i])
                        if verts[i] in remainingVerts:
                            remainingVerts.remove(verts[i])
                        break

            if [verts.index(sortedVerts[-1]), 0] in outlineEdges:
                cutEdges.append([verts.index(sortedVerts[-1]), 0])

            result.append(numpy.array(cutEdges))
            edgeCount = edgeCount + len(cutEdges)

        if edgeCount < len(verts):
            return None

        return result


    # create MeshData for _cutSN to visualize the cutting area as a square directly intersecting the model
    def drawCuttingArea(self):
        frontStartVert, frontEndVert, backStartVert, backEndVert = self.makeQuad(self._startVertex, self._endVertex, self._camRay)
        mb = MeshBuilder()
        mb.addQuad(
            v0 = frontStartVert,
            v1 = frontEndVert,
            v2 = backEndVert,
            v3 = backStartVert,
            color = Color(1.0, 0.0, 0.0, 0.5)
        )
        mb_data = mb.build()
        center = self.calcCenter(frontStartVert, frontEndVert, backEndVert,  backStartVert)
        self._cutSN.setCenterPosition(center)
        self._cutSN.setPosition(center, SceneNode.TransformSpace.Parent)
        self._cutSN.setMeshData(mb_data)

    # Helper Function for turning start and end point of mouse drag into a square in 3D-space
    # the mouse draw vector is "extruded" in the camera's viewing direction to create square
    def makeQuad(self, vert1: Vector, vert2: Vector, ray: Ray):
        side = numpy.subtract(vert2.getData(), vert1.getData())
        sideVec = Vector(side[0], side[1], side[2])
        #sideVec = self.getDragVector()
        dimension = sideVec.length() * 2

        StartRay = Ray(vert1, ray.direction)
        frontStartVert = StartRay.getPointAlongRay(-0.5 * dimension)
        backStartVert = StartRay.getPointAlongRay(0.5 * dimension)
        EndRay = Ray(vert2, ray.direction)
        frontEndVert = EndRay.getPointAlongRay(-0.5 * dimension)
        backEndVert = EndRay.getPointAlongRay(0.5 * dimension)

        return frontStartVert, frontEndVert, backStartVert, backEndVert

    # Helper Function for calculating center point of square in 3D-space
    def calcCenter(self, v0: Vector, v1: Vector, v2: Vector, v3: Vector) -> Vector:
        diag = numpy.subtract(v0.getData(), v2.getData())
        diagVec = v2 - v0
        diagRay = Ray(v0, diagVec)
        mid = diagRay.getPointAlongRay(0.5 * diagVec.length())
        return mid

    # Helper function for checking whether a given vertex lies within the plane
    def isVertexInPlane(self, vert, planeOrg, planeNrm):
        result = numpy.dot(numpy.subtract(planeOrg, vert), planeNrm)

        return Float.fuzzyCompare(result, 0.0, 0.000001)

    # Helper function, transforms list of given vertices on the same plane into 2D coordinates if they are 3D,
    # or into 3D coordinates if they are 2D, transformation matrix is passed as parameter M
    def transformToOtherDimension(self, verts, M):

        if numpy.shape(verts)[1] == 3:
            newVerts = numpy.zeros([numpy.shape(verts)[0], 2], dtype = numpy.float64)
            for i in range(numpy.shape(verts)[0]):
                v = numpy.dot(M, numpy.append(verts[i], [1.0]))
                newVerts[i][0] = v[0]
                newVerts[i][1] = v[1]
            return newVerts

        elif numpy.shape(verts)[1] == 2:
            newVerts = numpy.zeros([numpy.shape(verts)[0], 3], dtype = numpy.float64)
            for i in range(numpy.shape(verts)[0]):
                v = numpy.dot(M, numpy.append(verts[i], [0.0, 1.0]))
                newVerts[i][0] = v[0]
                newVerts[i][1] = v[1]
                newVerts[i][2] = v[2]
            return newVerts

        else:
            self._error.setText("There was a problem with the cutting process. (Error in transformDimension)")
            return None


    # Helper Function adapted from https://stackoverflow.com/questions/49769459/convert-points-on-a-3d-plane-to-2d-coordinates
    # Calculates transformation matrix for 3D to 2D vertices transformation
    # !! Matrices M and Minv are 4x4 !!
    def calc3Dto2DTransformMatrix(self, A, B, C):
        if self.areVectorsCollinear(A, B, C):
            self._error.setText("There was a problem with the cutting process. (Error in 3D-2D)")
            return None, None

        AB = numpy.subtract(B, A)
        AC = numpy.subtract(C, A)
        N = numpy.cross(AB, AC)
        Nn = N / numpy.linalg.norm(N)
        ABn = AB / numpy.linalg.norm(AB)
        V = numpy.cross(ABn, Nn)

        D = numpy.array([[0,1,0,0], [0,0,1,0], [0,0,0,1], [1,1,1,1]], dtype = numpy.float32)
        S = numpy.array([[A[0], A[0]+ABn[0], A[0]+V[0], A[0]+Nn[0]],
                         [A[1], A[1]+ABn[1], A[1]+V[1], A[1]+Nn[1]],
                         [A[2], A[2]+ABn[2], A[2]+V[2], A[2]+Nn[2]], [1,1,1,1]], dtype = numpy.float32)
        Sinv = numpy.linalg.inv(S)
        M = numpy.dot(D, Sinv)
        Minv = numpy.linalg.inv(M)

        return M, Minv

    # Helper function for finding three non-collinear vertices in list of vertices
    def getNonCollinearVertices(self, verts):
        # triple for loop for finding three collinear vertices
        for i in range(numpy.shape(verts)[0]):
            for j in range(i+1, numpy.shape(verts)[0]):
                for k in range(j+1, numpy.shape(verts)[0]):
                    if not self.areVectorsCollinear(verts[i], verts[j], verts[k]):
                        return verts[i], verts[j], verts[k]

        return None, None, None

    # Helper function for testing whether three given vertices are collinear
    def areVectorsCollinear(self, A, B, C) -> bool:
        AB = numpy.subtract(B, A)
        AC = numpy.subtract(C, A)
        nrm = numpy.cross(AB, AC)

        result = numpy.array_equal(nrm, numpy.array([0,0,0], dtype=numpy.float64))

        return result


    # Helper function taken from https://github.com/fieldOfView/Cura-MeshTools/blob/3.5/MeshTools.py
    # Transforms Uranium MeshData into python trimesh
    def _toTriMesh(self, mesh_data: MeshData) -> trimesh.base.Trimesh:
        indices = mesh_data.getIndices()
        if indices is None:
            # some file formats (eg 3mf) don't supply indices, but have unique vertices per face
            indices = numpy.arange(mesh_data.getVertexCount()).reshape(-1, 3)

        return trimesh.base.Trimesh(vertices=mesh_data.getVertices(), faces=indices, vertex_normals=mesh_data.getNormals())

    # Helper function taken from https://github.com/fieldOfView/Cura-MeshTools/blob/3.5/MeshTools.py
    # Transforms python trimesh into Uranium MeshData
    def _toMeshData(self, tri_node: trimesh.base.Trimesh, file_name: str = "") -> MeshData:
        tri_faces = tri_node.faces
        tri_vertices = tri_node.vertices

        indices = []
        vertices = []

        index_count = 0
        face_count = 0
        for tri_face in tri_faces:
            face = []
            for tri_index in tri_face:
                vertices.append(tri_vertices[tri_index])
                face.append(index_count)
                index_count += 1
            indices.append(face)
            face_count += 1

        vertices = numpy.asarray(vertices, dtype=numpy.float32)
        indices = numpy.asarray(indices, dtype=numpy.int32)
        normals = calculateNormalsFromIndexedVertices(vertices, indices, face_count)

        mesh_data = MeshData(file_name = file_name, vertices=vertices, indices=indices, normals=normals)
        return mesh_data


    # Helper function taken from https://github.com/fieldOfView/Cura-MeshTools/blob/3.5/MeshTools.py
    # given the seperate trimeshes calculated in splitMesh after a cut is performed, this function
    # makes the model being cut into two seperate, independent SceneNodes/objects
    def _replaceSceneNode(self, existing_node, trimeshes) -> None:
        name = existing_node.getName()
        file_name = existing_node.getMeshData().getFileName()
        transformation = existing_node.getWorldTransformation()
        parent = existing_node.getParent()
        extruder_id = existing_node.callDecoration("getActiveExtruder")
        build_plate = existing_node.callDecoration("getBuildPlateNumber")
        selected = Selection.isSelected(existing_node)

        children = existing_node.getChildren()
        new_nodes = []

        op = GroupedOperation()
        op.addOperation(RemoveSceneNodeOperation(existing_node))

        for i, tri_node in enumerate(trimeshes):
            mesh_data = self._toMeshData(tri_node, file_name)

            new_node = CuraSceneNode()
            new_node.setSelectable(True)
            new_node.setMeshData(mesh_data)
            new_node.setName(name if i==0 else "%s %d" % (name, i))
            new_node.callDecoration("setActiveExtruder", extruder_id)
            new_node.addDecorator(BuildPlateDecorator(build_plate))
            new_node.addDecorator(SliceableObjectDecorator())

            op.addOperation(AddSceneNodeOperation(new_node, parent))
            op.addOperation(SetTransformMatrixOperation(new_node, transformation))

            new_nodes.append(new_node)

            if selected:
                Selection.add(new_node)

        for child in children:
            child_bounding_box = child.getMeshData().getTransformed(child.getWorldTransformation()).getExtents()
            new_parent = None
            for potential_parent in new_nodes:
                parent_bounding_box = potential_parent.getMeshData().getTransformed(potential_parent.getWorldTransformation()).getExtents()
                if child_bounding_box.intersectsBox(parent_bounding_box) != AxisAlignedBox.IntersectionResult.NoIntersection:
                    new_parent = potential_parent
                    break
            if not new_parent:
                new_parent = new_nodes[0]
            op.addOperation(SetParentOperationSimplified(child, new_parent))

        op.push()
